import os
import json
import tempfile
from fractions import Fraction

from ..submission import BrokenSubmissionError
from ..utils import run_process, PIPE, STDOUT, TimeoutExpired
from ..grades import PartGrade
from . import GraderInterface, Part

"""
Grade a homework using the classic Java bitwise operators grader
(usually homework 2).
"""


class BitwiseJSONMethod(Part):
    __slots__ = ('class_name', 'method',)

    def __init__(self, class_name, method):
        self.class_name = class_name
        self.method = method

    def description(self):
        return '{}.{}()'.format(self.class_name, self.method)

    def grade(self, jsonResults):
        # If the method is missing from the results, they probably
        # didn't implement it.
        if self.method not in jsonResults:
            return PartGrade(score=Fraction(0), log='method not found',
                             deductions=('missing',))
        else:
            jsonResult = jsonResults[self.method]

        if jsonResult.get('errorMessage', None):
            return PartGrade(score=Fraction(0), log=jsonResult['errorMessage'])

        violations = jsonResult.get('violations', [])

        # If they violated any rules, boom, zero
        if violations:
            score = Fraction(0)
        else:
            score = Fraction(jsonResult['testsPassed'],
                             jsonResult['testsTotal'])

        log = jsonResult.get('message', '')

        return PartGrade(score=score, deductions=violations, log=log)


class BitwiseJSONGrader(GraderInterface):
    """
    Run the legendary hw2checker.jar and collect results.
    """

    DEFAULT_TIMEOUT = 10

    def __init__(self, grader_jar, source_file, timeout=None):
        self.grader_jar = grader_jar
        self.source_file = source_file

        if timeout is None:
            self.timeout = self.DEFAULT_TIMEOUT
        else:
            self.timeout = timeout

    def list_prerequisites(self):
        return ['openjdk-8-jre-headless']

    def class_name(self):
        """Return the class name of this file"""
        return self.source_file.rsplit('.', maxsplit=1)[0]

    def part_from_config_dict(self, config_dict):
        return BitwiseJSONMethod.from_config_dict(
            config_dict, class_name=self.class_name())

    def grade(self, submission, path, parts):
        gradelog_fp, gradelog_path = tempfile.mkstemp(prefix='log-',
                                                      suffix='.json', dir=path)
        # Don't leak fds
        os.close(gradelog_fp)

        cmdline = ['java', '-jar', self.grader_jar, '-z', self.source_file,
                   gradelog_path]
        try:
            process = run_process(cmdline, cwd=path, timeout=self.timeout,
                                  stdout=PIPE, stderr=STDOUT, input='')
        except TimeoutExpired:
            raise BrokenSubmissionError('timeout of {} seconds expired for '
                                        'grader'.format(self.timeout))

        if process.returncode != 0:
            raise BrokenSubmissionError(
                ('grader command exited with nonzero exit code {}. '
                 'syntax error?')
                .format(process.returncode),
                verbose=process.stdout.decode() if process.stdout else None)

        with open(gradelog_path) as gradelog_file:
            gradelog = json.load(gradelog_file)

        if gradelog.get('errorMessage', None):
            raise BrokenSubmissionError(gradelog['errorMessage'])

        return [part.grade(gradelog['results']) for part in parts]



from abc import ABCMeta, abstractmethod

from ..utils import ConfigDictMixin


class PenalizerInterface(ConfigDictMixin):
    """Penalize a student outside component grades. Example: Late penalties"""

    __metaclass__ = ABCMeta

    def __init__(self):
        """
        The class needs an init method that will take in all of its
        desired options from the config file as keyword args. Required
        options that need to be in the config file should not have
        default values. Optional ones can have default values but still
        need to exist as kwargs. Each of the elements in the options:
        section of the component config of the assignment will be passed
        as a keyword argument during initialization.
        """
        pass

    @abstractmethod
    def adjust_grade(self,
                     submission,
                     grade):
        """
        This function should take in a Submission object and a Fraction
        holding the calculated grade, returning the new grade.
        """
        pass



import sys
import os.path
from importlib import import_module
from fractions import Fraction

from ..submission import BrokenSubmissionError
from ..grades import PartGrade
from . import GraderInterface, Part

"""
Grade a homework using a custom python module in grading-files/. This
way you don't need to exec() anything and can grade a homework from the
comfort of your own cozy Python function.
"""


class PythonModulePart(Part):
    __slots__ = ('id', 'name')

    def __init__(self, id, name):
        self.id = id
        self.name = name

    def description(self):
        return self.name

    def grade(self, result):
        if result is None:
            score = Fraction(0)
            log = "missing part id in result!"
        elif isinstance(result, Exception):
            score = Fraction(0)
            log = 'grading part threw exception: {}'.format(result)
        # Should be a fractions.Fraction
        else:
            score = Fraction(result)
            log = ''

        return PartGrade(score=score, log=log)


class PythonModuleGrader(GraderInterface):
    """
    Run a Python module and collect the results. Very basic, will
    probably need to be expanded down the road.

    The module needs to be in grading-files/, and the function set by
    `function' in zucchini.yml needs to return a dict mapping part_id ->
    fractions.Fraction (or -> Exception if something went wrong).
    """

    def __init__(self, module, function, student_file):
        self.module = module
        self.function = function
        self.student_file = student_file

    def part_from_config_dict(self, config_dict):
        return PythonModulePart.from_config_dict(config_dict)

    def grade(self, submission, path, parts):
        old_sys_path = sys.path[:]
        sys.path.append(path)

        try:
            return self._grade(submission, path, parts)
        finally:
            sys.path = old_sys_path

    def _grade(self, submission, path, parts):
        try:
            autograder_module = import_module(self.module)
        except Exception as err:
            raise BrokenSubmissionError("error importing module `{}': {}"
                                        .format(self.module, str(err)))

        autograder_func = getattr(autograder_module, self.function)

        student_file_path = os.path.join(path, self.student_file)

        try:
            results = autograder_func(student_file_path)
        except Exception as err:
            raise BrokenSubmissionError("autograder function choked: {}"
                                        .format(str(err)))

        return [part.grade(results.get(part.id)) for part in parts]



"""Interface with the Canvas through its REST API"""

import os
import re
import requests
import shutil
from collections import namedtuple

from .utils import datetime_from_string

# json.load() started throwing JSONDecodeError instead of ValueError
# back in Python 3.5, so catch ValueError instead for earlier versions
try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


class CanvasAPIError(Exception):
    """Raised for bad responses to API requests"""
    pass


class CanvasNotFoundError(CanvasAPIError):
    """Raised when a request 404s"""
    pass


class CanvasInternalError(CanvasAPIError):
    """Raised when a response is a 5xx"""
    pass


class CanvasMalformedResponseError(CanvasAPIError):
    """Raised when a response contains invalid JSON"""
    pass


class CanvasCourse(namedtuple('CanvasCourse', ('api_', 'id', 'name'))):
    """Hold Course info"""
    __slots__ = ()

    def __str__(self):
        return 'id={}\t{}'.format(self.id, self.name)

    @property
    def assignments(self):
        return self.api_.list_assignments(self.course_id)

    @property
    def sections(self):
        return self.api_.list_sections(self.course_id)


class CanvasAssignment(namedtuple('CanvasAssignment',
                       ('api_', 'id', 'name', 'course_id'))):
    """Hold assignment info"""
    __slots__ = ()

    def __str__(self):
        return 'id={}\t{}'.format(self.id, self.name)

    @property
    def course(self):
        return self.api_.get_course(self.course_id)


class CanvasSection(namedtuple('CanvasSection',
                    ('api_', 'id', 'name', 'course_id'))):
    """Hold section info"""
    __slots__ = ()

    def __str__(self):
        return 'id={}\t{}'.format(self.id, self.name)

    @property
    def course(self):
        return self.api_.get_course(self.course_id)


class CanvasUser(namedtuple('CanvasUser',
                 ('api_', 'id', 'name', 'sortable_name'))):
    """Hold user info"""
    __slots__ = ()

    def __str__(self):
        return 'id={}\t{}'.format(self.id, self.name)


class CanvasSubmissionAttachment(namedtuple('CanvasSubmission',
                                 ('api_', 'id', 'filename', 'url'))):
    """Hold information about a submitted file"""
    __slots__ = ()

    def __str__(self):
        return 'id={}\t{}\t{}' \
               .format(self.id, self.filename, self.url)

    def download(self, directory):
        """
        Download this attachment with the name provided by the student
        to the directory provided.
        """
        # Canvas sanitizes the input filename, it seems
        path = os.path.join(directory, self.filename)
        self.api_._download_file(self.url, path)


class CanvasSubmission(namedtuple('CanvasSubmission',
                       ('api_', 'id', 'late', 'user_id', 'user', 'attachments',
                        'seconds_late', 'attempt'))):
    """Hold assignment info"""
    __slots__ = ()
    _defaults = {'attachments': []}
    _sub_entities = {'user': CanvasUser}
    _sub_entity_lists = {'attachments': CanvasSubmissionAttachment}

    def __str__(self):
        attachments = ', '.join(attachment.filename
                                for attachment in self.attachments)
        return 'id={}\t{}late\t{}\t{}' \
               .format(self.id, '' if self.late else 'not ', self.user,
                       attachments)

    def no_submission(self):
        """Return True iff the student did not submit."""
        return self.attempt is None

    def time(self):
        """Return the submission time as a naive datetime object."""
        if self.submitted_at is None:
            return self.submitted_at
        else:
            return datetime_from_string(self.submitted_at)

    def download(self, directory):
        """
        Download all the attachments for this submission to the given
        directory.
        """
        for attachment in self.attachments:
            attachment.download(directory)


class CanvasAPI(object):
    """
    Represent a connection to the Canvas REST API.

    Currently does not support OAuth2 because Canvas OAuth2 would be a
    mess with Zucchini. We'd have to register it as a "Developer Key"
    with OIT and then pass around a `client_id' and `client_secret'.
    Much easier to be a little shady with the terms of service and get
    'er done.
    """

    LINK_REGEX = re.compile(r'<(?P<link>.+?)>; rel="(?P<rel>.+?)"(,|$)')

    def __init__(self, url, token):
        self.url = url.rstrip('/')
        self.token = token

    def _parse_links(self, link_header):
        """
        Convert the contents of a link header into a dictionary of
        rel->link. See
        https://canvas.instructure.com/doc/api/file.pagination.html for
        more information.
        """

        return {match.group('rel'): match.group('link')
                for match in self.LINK_REGEX.finditer(link_header)}

    def _to_entity(self, entity_json, entity_class):
        """Convert a JSON object to an instance of the provided entity class"""

        if hasattr(entity_class, '_defaults'):
            defaults = entity_class._defaults
        else:
            defaults = {}

        args = {}
        for field in entity_class._fields:
            # Ignore the special api_ field used to give the instance a
            # back-reference to this class
            if field == 'api_':
                continue

            if field in entity_json:
                args[field] = entity_json[field]
            elif field in defaults:
                args[field] = defaults[field]
            else:
                raise CanvasMalformedResponseError(
                    'response is missing required field {}'.format(field))

        if hasattr(entity_class, '_sub_entities'):
            for field, class_ in entity_class._sub_entities.items():
                args[field] = self._to_entity(args[field], class_)

        if hasattr(entity_class, '_sub_entity_lists'):
            for field, class_ in entity_class._sub_entity_lists.items():
                args[field] = [self._to_entity(entity, class_)
                               for entity in args[field]]

        return entity_class(api_=self, **args)

    def _url(self, name):
        """Return the URL of an endpoint named `name'"""
        return '{}/api/v1/{}'.format(self.url, name)

    def _headers(self):
        """Return authentication headers"""
        return {'authorization': 'Bearer {}'.format(self.token)}

    def _request(self, url, method='get', json=None, stream=False):
        """
        Attempt to request the URL url, raising exceptions in error
        conditions
        """
        resp = requests.request(method, url, headers=self._headers(),
                                json=json, stream=stream)

        if resp.status_code == 404:
            raise CanvasNotFoundError()
        elif resp.status_code >= 500 and resp.status_code < 600:
            raise CanvasInternalError()
        elif resp.status_code != 200:
            raise CanvasAPIError('unexpected response code {}'
                                 .format(resp.status_code))
        return resp

    def _put(self, name, json):
        self._request(self._url(name), method='put', json=json)

    def _get_json(self, name):
        """
        Make a request to a Canvas API endpoint and return the JSON.
        """
        try:
            return self._request(self._url(name)).json()
        except JSONDecodeError as err:
            raise CanvasMalformedResponseError(err)

    def _get(self, name, entity_class):
        """
        Make a request to a Canvas API endpoint and convert the JSON
        entity in the response to an instance of `entity_class'.
        """

        return self._to_entity(self._get_json(name), entity_class)

    def _gets(self, name, entity_class):
        """
        Make a request to a Canvas API endpoint, de-paginate the
        response, and convert each JSON entity to an instance of
        `entity_class'.
        """

        link = self._url(name)

        while link:
            resp = self._request(link)

            if 'link' in resp.headers:
                link = self._parse_links(resp.headers['link']) \
                           .get('next', None)
            else:
                link = None

            try:
                json = resp.json()
            except JSONDecodeError as err:
                raise CanvasMalformedResponseError(err)

            for entity_json in json:
                yield self._to_entity(entity_json, entity_class)

    def _download_file(self, url, path):
        """Download the file at `url' to `path'."""

        with self._request(url, stream=True) as response, \
                open(path, 'wb') as downloaded_file:
            shutil.copyfileobj(response.raw, downloaded_file)

    def list_courses(self):
        """
        Return an iterator of CanvasCourse instances of all the active
        courses this user can see.
        """

        return self._gets('courses', CanvasCourse)

    def get_course(self, course_id):
        """
        Return a CanvasCourse instance corresponding to the given
        course id.
        """

        return self._get('courses/{}'.format(course_id), CanvasCourse)

    def get_user(self, course_id):
        """
        Return a CanvasCourse instance corresponding to the given
        course id.
        """

        return self._get('users/{}'.format(course_id), CanvasUser)

    def list_assignments(self, course_id):
        """
        Return an iterator of instances of all the assignments in the
        given course.
        """

        return self._gets('courses/{}/assignments'.format(course_id),
                          CanvasAssignment)

    def list_sections(self, course_id):
        """
        Return an iterator of instances of all the sections in the
        given course.
        """

        return self._gets('courses/{}/sections'.format(course_id),
                          CanvasSection)

    def list_section_students(self, course_id, section_id):
        """
        Return an iterator of instances of all the sections in the
        given course.
        """

        # XXX sections/{}?include=students crashes Canvas, so avoid for now
        section = self._get_json('courses/{}/sections/{}?include=students'
                                 .format(course_id, section_id))

        for student_json in section['students']:
            yield self._to_entity(student_json, CanvasUser)

    def list_submissions(self, course_id, assignment_id):
        """
        Return an iterator of instances of all the submissions in the
        given course for the given assignment.
        """

        return self._gets('courses/{}/assignments/{}/submissions?include=user'
                          .format(course_id, assignment_id), CanvasSubmission)

    def list_section_submissions(self, section_id, assignment_id):
        """
        Return an iterator of instances of all the submissions in the
        given section for the given assignment.
        """

        return self._gets('sections/{}/assignments/{}/submissions?include=user'
                          .format(section_id, assignment_id), CanvasSubmission)

    def get_submission(self, course_id, assignment_id, user_id):
        """
        Return a CanvasSubmission instance for the given user in the
        given assignment.
        """
        return self._get(('courses/{}/assignments/{}/submissions/{}'
                         '?include=user')
                         .format(course_id, assignment_id, user_id),
                         CanvasSubmission)

    def set_submission_grade(self, course_id, assignment_id, user_id, grade,
                             comment=None):
        """
        Set the grade and optionally add a comment to the submission for
        the given user in the given assignment.

        Warning: specifying `comment' will add a new comment, not edit
        an existing one. So every time you call this method with
        comment='XYZ', you will add a new comment.
        """

        json = {'submission': {'posted_grade': '{}%'.format(grade)},
                'comment': {}}

        if comment is not None:
            json['comment']['text_comment'] = comment

        self._put('courses/{}/assignments/{}/submissions/{}'
                  .format(course_id, assignment_id, user_id), json)

    def add_submission_comment(self, course_id, assignment_id, user_id,
                               comment, files):
        """
        Adds a comment with optional file(s) attached to a submission

        Info on uploading a file for comments. Warning! it's a very
        involved process
        https://canvas.instructure.com/doc/api/submission_comments.html

        :param course_id: int canvas course id
        :param assignment_id: int canvas assignment id
        :param user_id: int canvas user_id
        :param comment: string comment that will be added to submission
        :param files: (optional) List of (paths to files, file mime type)
        tuples that will be attached to comment
        :return:

        wow wouldn't it be great if you could auto detect the mime type?
        yeah that would be great but windows doesn't have the binary
        needed for python-magic to work
        """

        session = requests.Session()
        session.headers = self._headers()

        file_ids = []

        for file in files or []:

            # Step 1: tell canvas about file upload
            file_path, mime_type = file
            part_1_dict = {
                'name': os.path.basename(file_path),
                'size': os.stat(file_path).st_size,
                'content_type': mime_type
            }

            part_1_resp = session.post(
                self._url('courses/%s/assignments/%s/submissions/%s/comments/'
                          'files' % (course_id, assignment_id, user_id)),
                json=part_1_dict
            )

            part_1_resp.raise_for_status()

            # Step 2: Upload the file data to the URL given in the
            # previous response

            # construct form
            with open(file_path, 'rb') as f:
                file_contents = f.read()
            upload_params = list(part_1_resp.json()['upload_params'].items())
            upload_params.append(('file', file_contents))

            part_2_resp = requests.post(
                part_1_resp.json()['upload_url'],
                files=upload_params
            )

            part_2_resp.raise_for_status()

            file_ids.append(part_2_resp.json()['id'])

        # make the comment
        part_3_req = {
            'comment': {
                'text_comment': comment,
                'file_ids': file_ids
            }
        }
        part_3_resp = session.put(
            self._url('courses/%s/assignments/%s/submissions/%s' %
                      (course_id, assignment_id, user_id)),
            json=part_3_req
        )

        part_3_resp.raise_for_status()



import re
from fractions import Fraction

from . import PenalizerInterface, InvalidPenalizerConfigError
from ..utils import ConfigDictMixin

"""Penalize a late submission."""


class LatePenalty(ConfigDictMixin):
    UNITS_REGEX = re.compile(r'^(?P<mag>[0-9/]+)\s*(?P<unit>[a-z]*)$')

    def __init__(self, after, penalty):
        self.after = self.time_to_seconds(after)
        self.penalty, penalty_unit = self.split_units(penalty)

        if penalty_unit in ('pt', 'pts'):
            self.penalty /= 100
            self.penalty_points = True
        elif penalty_unit is None:
            self.penalty_points = False
        else:
            raise InvalidPenalizerConfigError("unknown penalty unit `{}'. try "
                                              "a fraction optionally followed "
                                              "by `pt'."
                                              .format(penalty_unit))

    @classmethod
    def split_units(cls, amount_str):
        if isinstance(amount_str, (int, float)):
            return Fraction(amount_str), None

        match = cls.UNITS_REGEX.match(amount_str.lower())

        if match is None:
            raise InvalidPenalizerConfigError("unknown units format `{}'"
                                              .format())

        return Fraction(match.group('mag')), match.group('unit') or None

    @classmethod
    def time_to_seconds(cls, time_str):
        mag, unit = cls.split_units(time_str)

        unit = unit or 's'
        units = {'s': 1, 'm': 60, 'h': 60*60, 'd': 24*60*60}

        if unit not in units:
            raise InvalidPenalizerConfigError("unknown time unit `{}'. try "
                                              "one of {}"
                                              .format(unit, ', '.join(units)))
        return mag * units[unit]

    def is_late(self, submission):
        return submission.seconds_late is not None \
               and submission.seconds_late > self.after

    def adjust_grade(self, grade):
        if self.penalty_points:
            return max(0, grade - self.penalty)
        else:
            return grade * (1 - self.penalty)


class LatePenalizer(PenalizerInterface):
    """
    Penalize students for late submissions.

    Configure it like this in the assignment config file. In this
    example, after 8 hours late you get 75 points off your grade (a 85
    would go to a 10).

    penalties:
    - name: LATE
      backend: LatePenalizer
      backend-options:
        penalties:
        - after: 1h
          penalty: 25pts
        - after: 8h
          penalty: 50pts

    That is, penalties are applied in order, and they do not stop when
    there is a match. They don't necessarily have to be in increasing
    order of `after' values.
    """

    def __init__(self, penalties):
        self.penalties = [LatePenalty.from_config_dict(p) for p in penalties]

    def adjust_grade(self, submission, grade):
        for penalty in self.penalties:
            if penalty.is_late(submission):
                grade = penalty.adjust_grade(grade)

        return grade



from .constants import READ_BUFFER_SIZE
import hashlib
import os
import boto3


class AmazonAPI(object):
    """
    Represent a connection to the AWS API.
    """

    def __init__(self, access_key_id, secret_access_key, bucket_name):
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.bucket_name = bucket_name

    def upload_file_s3(self, file_path, file_mime_type):

        client = boto3.client(
            's3',
            aws_access_key_id=self.access_key_id,
            aws_secret_access_key=self.secret_access_key)

        file_name = os.path.basename(file_path)
        mid_idx = file_name.rfind('.')
        file_sha256 = AmazonAPI._calculate_hash(file_path, 'sha256')
        s3_file_name = '%s-%s%s' % (
            file_name[0:mid_idx],
            file_sha256[0:15],
            file_name[mid_idx:])

        client.put_object(
            Bucket=self.bucket_name,
            ACL='public-read',
            ContentType=file_mime_type,
            Key=s3_file_name,
            Body=open(file_path, 'rb'))

        return 'https://%s.s3.amazonaws.com/%s' % \
               (self.bucket_name, s3_file_name)

    @classmethod
    def _calculate_hash(cls, file_path, algo):
        h = hashlib.new(algo)
        with open(file_path, 'rb') as f:
            while True:
                d = f.read(READ_BUFFER_SIZE)
                if not d:
                    break
                h.update(d)
        return h.hexdigest()



import shlex
from fractions import Fraction

from ..submission import BrokenSubmissionError
from ..utils import run_process, PIPE, STDOUT, TimeoutExpired
from ..grades import PartGrade
from . import GraderInterface, Part

"""
Grade a homework by execuing a bunch of commands
"""


class Command(Part):
    """A command to execute"""

    __slots__ = ('summary', 'command')

    def __init__(self, summary, command):
        self.summary = summary
        self.command = shlex.split(command)

    def description(self):
        return self.summary

    def grade(self, path, timeout):
        try:
            process = run_process(self.command, cwd=path, timeout=timeout,
                                  stdout=PIPE, stderr=STDOUT, input='')
        except TimeoutExpired:
            raise BrokenSubmissionError('timeout of {} seconds expired'
                                        .format(self.timeout))

        if process.stdout is None:
            log = '(no output)'
        else:
            log = process.stdout.decode()

        if process.returncode:
            score = Fraction(0)
            log += '\n\nprocess exited with exit code {} != 0' \
                   .format(process.returncode)
        else:
            score = Fraction(1)

        return PartGrade(score=score, log=log)


class MultiCommandGrader(GraderInterface):
    """
    Run a bunch of commands, testing the exit code of each
    """

    DEFAULT_TIMEOUT = 30

    def __init__(self, timeout=None, extra_setup_commands=None):
        if timeout is None:
            self.timeout = self.DEFAULT_TIMEOUT
        else:
            self.timeout = timeout

        if extra_setup_commands is None:
            self.extra_setup_commands = []
        else:
            self.extra_setup_commands = extra_setup_commands

    def list_extra_setup_commands(self):
        return self.extra_setup_commands

    def part_from_config_dict(self, config_dict):
        return Command.from_config_dict(config_dict)

    def grade(self, submission, path, parts):
        return [part.grade(path, self.timeout) for part in parts]



"""
Utilities for local autograding.
"""


class LocalAutograderOutput:
    """
    Take a Grade object and convert that to an output string that will be
    printed into the terminal
    """

    @classmethod
    def from_grade(cls, grade):
        computed_grade = grade.computed_grade()
        output = []
        for component in computed_grade.components:
            if component.error:
                output.append("ERROR: %-45s %-15s" %
                              (component.name, component.error))
                output.append("Details: %-45s" % component.error_verbose)
            else:
                for part in component.parts:
                    if part.points_got < part.points_possible:
                        output.append("TEST: %-45s %-15s" %
                                      (part.name, "FAILED"))
                        output.append(part.log)
                    else:
                        output.append("TEST: %-45s %-15s" %
                                      (part.name, "PASSED"))
        return '\n\n'.join(output)



import os
import re
import shutil
from zipfile import ZipFile

from ..utils import sanitize_path, mkdir_p


class CanvasArchiveLoader(object):
    FILENAME_REGEX = re.compile(r'^(?P<user_name>[a-z\-_]+)(?P<user_id>\d+)_'
                                r'question_\d+_\d+_(?P<filename>.+)$')

    def __init__(self, zipfile_path):
        self.zipfile_path = zipfile_path
        self.zipfile = None
        # Map from student id to list of (zip_filename, original_filename)
        self.submissions = {}

    def __enter__(self):
        self.zipfile = ZipFile(self.zipfile_path)
        self.load()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.zipfile.close()

    def has_submission(self, user_id):
        return user_id in self.submissions

    def load(self):
        for filename in self.zipfile.namelist():
            match = self.FILENAME_REGEX.match(filename)
            if match is None:
                raise Exception('filename in archive in invalid format')
            self.submissions.setdefault(int(match.group('user_id')), []) \
                .append((filename, match.group('filename')))

    def extract_files(self, user_id, dest_dir):
        for zip_filename, filename in self.submissions.get(user_id, []):
            fullpath = os.path.join(dest_dir, sanitize_path(filename))
            mkdir_p(os.path.dirname(fullpath))

            with self.zipfile.open(zip_filename) as sourcefile, \
                    open(fullpath, 'wb') as destfile:
                shutil.copyfileobj(sourcefile, destfile)



from abc import ABCMeta, abstractmethod


class ExporterInterface:
    __metaclass__ = ABCMeta

    # The class needs an init method that will take in all of its desired
    # options from the command line, in order.

    @abstractmethod
    def export(self, submission):
        """This method should take in a Submission object and export its
        grading to whatever export it's doing.
        For example, if we're exporting to CSV, the Exporter object should
        have the CSV file as an attribute, and
        this function should write one submission's grade to the CSV file."""
        pass

    def __enter__(self):
        """This method allows us to use the with keyword for our exporter."""
        return self

    def __exit__(self, *args):
        """This method will be used to close any open files etc. at the end of
        the exporting procedure.
        It is automatically called when the with block terminates. Close any
        files and complete any submissions here."""
        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import sys
from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'click==6.7',
    'gitdb2==2.0.3',
    'GitPython==2.1.8',
    'PyYAML==3.13',
    'smmap2==2.0.3',
    'requests==2.18.4',
    'boto3==1.5.19',
    'subprocess32==3.2.7;python_version<"3.0"',
    'arrow==0.12.1',
]

setup_requirements = [
    # TODO(zucchini): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='zucchini',
    version='1.0.1',
    description="Zucchini is an automatic grader tool for use in grading programming assignments.",
    long_description=readme + '\n\n' + history,
    author="Zucchini Team",
    author_email='team@zucc.io',
    url='https://github.com/zucchini/zucchini',
    packages=find_packages(include=['zucchini', 'zucchini.*']),
    entry_points={
        'console_scripts': [
            'zucc=zucchini.cli:cli'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='zucchini',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



import os
import decimal
import hashlib
import json
from fractions import Fraction

from .submission import Submission
from .constants import SUBMISSION_GRADELOG_FILE


def grade_all(submission):
    return True


class Grade(object):
    """
    Hold information about a student's grade and format grade
    breakdowns. Abstracts away the more annoying assignment grading
    logic from the command-line interface.
    """

    # Round 0.5->1 when rounding fraction scores
    ROUNDING = decimal.ROUND_HALF_UP

    # constants for gradelog files
    ZUCCHINI_BEGIN_GRADELOG = 'ZUCCHINI_BEGIN_GRADELOG'
    ZUCCHINI_END_GRADELOG = 'ZUCCHINI_END_GRADELOG'
    READ_BUFFER_SIZE = 65536

    def __init__(self, assignment, submission, max_score=100):
        self._assignment = assignment
        self._submission = submission
        self._max_score = max_score
        self._component_grades = self._submission.component_grades
        self._grade = None

    def __repr__(self):
        return '<Grade assignment={}, submission={}, component_grades={}, ' \
               'grade={}>'.format(self._assignment, self._submission,
                                  self._component_grades, self._get_grade())

    def gradable(self):
        """Return True if this submission is actually gradeable"""
        return self._submission.error is None

    def graded(self):
        """Return True if this submission has been graded, False otherwise."""
        return self._submission.graded

    def grade_ready(self):
        """Return True when this submission has a grade ready."""
        return not self.gradable() or self.graded()

    def update(self, grade):
        """
        Copy the non-None component grades of grade into this Grade.
        """
        for i, component_grade in enumerate(grade._component_grades):
            if component_grade is not None:
                self._component_grades[i] = component_grade

    def grade(self, interactive=None):
        """
        Grade the submission and calculate the grade.
        If interactive is None (the default), grade all components; if
        True, grade only interactive components, and if False, grade
        only noninteractive components.
        """
        self._component_grades = self._assignment.grade_submission(
                self._submission, interactive=interactive)

    def _get_grade(self):
        """
        Calculate the grade for this submission, or return it if already
        calculated.
        """
        if self._grade is not None:
            return self._grade
        elif self._component_grades is None:
            return None
        else:
            self._grade = self._assignment.calculate_grade(
                self._submission, self._component_grades)
            return self._grade

    def get_gradelog_path(self):
        # Put in the same directory as the metadata file
        submission_dir = os.path.dirname(self._submission.metadata_path)
        return os.path.join(submission_dir, SUBMISSION_GRADELOG_FILE)

    def get_gradelog_hash(self):
        with open(self.get_gradelog_path(), 'r') as f:
            gradelog_data = f.read()
            idx = gradelog_data.index(
                self.ZUCCHINI_END_GRADELOG) + len(self.ZUCCHINI_END_GRADELOG)
            return gradelog_data[idx:].strip()

    def serialized_component_grades(self):
        """
        Return component grades serialized with to_config_dict(). Useful
        for logging.
        """
        return [component_grade.to_config_dict()
                for component_grade in self._component_grades]

    def dump_component_grades(self, fp):
        """Dump component grades as JSON to file-like object fp"""
        json.dump(self.serialized_component_grades(), fp)

    def write_grade(self):
        """Write this grade to the submission metadata json."""
        # Need to put the components in the form used in the submission
        # meta.json
        self._submission.write_grade(self.serialized_component_grades())

    def student_name(self):
        """Return the name of the student."""
        return self._submission.student_name

    def student_id(self):
        """Return the id of the student, or None if unset."""
        return self._submission.id

    def to_float(self, frac):
        """
        Convert a fraction on [0,1] to a float to two decimal points.
        Should be used for ballparking, NOT calculations.
        """
        # We want a number on [0,100], not [0,1]
        out_of_max = frac * self._max_score
        return float(out_of_max.numerator) / float(out_of_max.denominator)

    def _decimal_out_of_max(self, frac):
        """
        Convert frac to a decimal.Decimal instance representing the
        score it represents out of 100 (or whatever the max score is).
        """
        # We want a number on [0,100], not [0,1]
        out_of_max = frac * self._max_score
        return decimal.Decimal(out_of_max.numerator) \
            / decimal.Decimal(out_of_max.denominator)

    def to_integer(self, frac):
        """
        Round frac to an integer out of 100 (or whatever the max score
        is)
        """
        quotient = self._decimal_out_of_max(frac)
        return int(quotient.to_integral_value(self.ROUNDING))

    def two_decimals(self, frac):
        """
        Convert frac to a string holding the number of points out of 100
        (or whatever the max score is) to two decimal points.
        """
        quotient = self._decimal_out_of_max(frac)
        # Round to two decimal places
        return str(quotient.quantize(decimal.Decimal('1.00'), self.ROUNDING))

    def _left_pad(self, num):
        return "%*.2f" % (5, num * self._max_score)

    def computed_grade(self):
        """
        Return a ComputedGrade instance holding the tree of computed
        grading data for this grade.
        """
        return self._get_grade()

    def score(self):
        """
        Return the grade as an integer out of 100 (or whatever the
        max score is).
        """
        grade = self._get_grade()
        if grade is None:
            return 0
        else:
            return self.to_integer(grade.grade)

    def _breakdown_deductions(self, grade):
        deducted_parts = []

        for penalty in grade.penalties:
            if penalty.points_delta != 0:
                sign = '+' if penalty.points_delta > 0 else ''
                points = self.two_decimals(penalty.points_delta)
                deducted_parts.append('{}: {}{}'.format(penalty.name, sign,
                                                        points))

        for component in grade.components:
            if component.error is not None:
                # Component-level error
                points = self.two_decimals(component.points_delta)
                deducted_parts.append('{}: {} ({})'.format(
                    component.name, points, component.error))
            else:
                for part in component.parts:
                    # To keep the breakdown easy to read, exclude parts
                    # which the student passed
                    if part.grade == 1:
                        continue

                    delta = self.two_decimals(part.points_delta)

                    if part.deductions:
                        deductions = ' ({})'.format(
                            ', '.join(part.deductions))
                    else:
                        deductions = ''

                    plus_sign = '+' if part.grade > 1 else ''
                    deducted_parts.append(
                        '{}: {}{}{}'.format(part.name, plus_sign,
                                            delta, deductions))

        return deducted_parts

    def breakdown(self, grader_name):
        """
        Generate a grade breakdown for this grade. Each part whose score
        != 1 is included.
        """

        if self.gradable():
            deductions = self._breakdown_deductions(self._get_grade())
            breakdown = ', '.join(deductions)
        else:
            breakdown = 'error: ' + self._submission.error

        return '{} -{}'.format(breakdown or 'Perfect!', grader_name)

    def _files_tuple_mapper(self, common_path, x):
        file_path = x[0]
        file_path_str = x[1]

        start_idx = file_path_str.find(common_path) + len(common_path)
        file_path_str_abbrv = file_path_str[start_idx:]

        file_hash = x[3]
        return (file_path, file_path_str, file_path_str_abbrv, file_hash)

    def generate_gradelog(self):
        """
        Writes gradelog file to submission directory
        """

        grade = self._get_grade()
        gradelog_path = self.get_gradelog_path()

        with open(gradelog_path, 'w') as f:
            # this used to be student's name, but that might be FERPA
            gradelog_path_byte_encoded \
                = str(self.get_gradelog_path()).encode('utf-8')
            f.write("%s\n\n%s\nstudent_name: \"%s\"" % (
                self.ZUCCHINI_BEGIN_GRADELOG,
                self._assignment.name,
                hashlib.sha224(gradelog_path_byte_encoded).hexdigest()[0:31]))

            if self._submission.seconds_late is not None:
                m, s = divmod(self._submission.seconds_late, 60)
                h, m = divmod(m, 60)

                f.write(", hours_late: %s" % "%d:%02d:%02d" % (h, m, s))

            f.write("\n\n")

            assignment_pass = True

            for component in grade.components:
                if component.error is not None:
                    assignment_pass = False
                    f.write("(%s / %s) [FAIL] TOTAL for %s: %s%s\n\n" % (
                        self._left_pad(component.points_got),
                        self._left_pad(component.points_possible),
                        component.name, component.error,
                        ('\n' + component.error_verbose)
                        if component.error_verbose else ''))
                else:
                    component_pass = True

                    for part in component.parts:
                        if part.grade >= 1:
                            f.write("(%s / %s) [PASS] %s: %s\n" % (
                                self._left_pad(part.points_got),
                                self._left_pad(part.points_possible),
                                component.name, part.name))
                        else:
                            component_pass = False
                            assignment_pass = False
                            f.write("(%s / %s) [FAIL] %s: %s - %s\n" % (
                                self._left_pad(part.points_got),
                                self._left_pad(part.points_possible),
                                component.name, part.name, part.log))

                    # print totals for assignment component
                    f.write("(%s / %s) [%s] TOTAL for %s\n\n" % (
                        self._left_pad(component.points_got),
                        self._left_pad(component.points_possible),
                        "PASS" if component_pass else "FAIL",
                        component.name))

            # print totals for complete assignment
            f.write("(%s / %s) [%s] TOTAL for %s (without penalties)\n" % (
                self._left_pad(grade.raw_grade),
                self._left_pad(Fraction(1)),
                "PASS" if assignment_pass else "FAIL",
                self._assignment.name))

            # print penalties (like being late)
            for penalty in grade.penalties:
                if penalty.points_delta != 0:
                    f.write("(%s) Penalty: %s\n" % (
                        self._left_pad(penalty.points_delta),
                        penalty.name
                    ))

            f.write("\n-----------------------\n| %s%% FINAL SCORE "
                    "|\n-----------------------\n\n" %
                    (self._left_pad(grade.grade)))

            # write filenames and hashes
            file_hashes, submission_hash = self.generate_submission_hash()
            f.write("---- File Hashes ----\n")
            # x is file_path and y is file_path_str, not used
            for x, y, file_path_str_abbrev, file_hash in file_hashes:
                f.write("(sha1: %s) %s\n" % (file_hash, file_path_str_abbrev))
            f.write("\n---- Submission Hash ----\n(sha1: %s)\n\n"
                    % submission_hash)
            f.write('%s\n' % self.ZUCCHINI_END_GRADELOG)

        # write gradelog hash
        with open(gradelog_path, 'r+') as f:
            file_data = f.read()
            begin_idx = file_data.index(self.ZUCCHINI_BEGIN_GRADELOG)
            end_idx = file_data.index(self.ZUCCHINI_END_GRADELOG)
            gradelog_data = file_data[begin_idx:end_idx]
            gradelog_hash = hashlib.sha1(gradelog_data.encode()).hexdigest()
            f.write("%s\n" % (gradelog_hash))

    def generate_submission_hash(self):
        files = []

        # get all files in submission directory
        submission_files_path = self._submission.files_path
        for (dirpath, dirnames, filenames) in os.walk(submission_files_path):
            for file_name in filenames:
                file_path = os.path.join(dirpath, file_name)

                # reads in file in chunks and gets hash
                hasher = hashlib.sha1()
                with open(file_path, 'rb') as f:
                    file_chunk = f.read(self.READ_BUFFER_SIZE)
                    if not file_chunk:
                        break
                    hasher.update(file_chunk)
                file_hash = hasher.hexdigest()

                files.append(
                    (file_path, str(file_path), "abbreviated path", file_hash))

        if len(files) < 1:
            return [], hashlib.sha1(''.encode('utf-8')).hexdigest()

        common_path = str(os.path.commonpath([x[0] for x in files]))

        files2 = []
        for x in files:
            files2.append(self._files_tuple_mapper(common_path, x))

        files2.sort(key=lambda x: x[2])
        files = files2

        # get submission hash (hash of all file hashes, in order)
        submission_hasher = hashlib.sha1()
        for file_path, file_path_str, file_path_str_abbrv, file_hash in files:
            submission_hasher.update(file_hash.encode('utf-8'))
        submission_hash = submission_hasher.hexdigest()

        return files, submission_hash


class GradingManager(object):
    def __init__(self, assignment, submission_path, filter_fn=grade_all):
        # Here we've abstracted away the submission filter. If the user wants
        # to selectively grade some submissions, the CLI will provide a filter
        # function that, when given a submission's name, returns a boolean
        # denoting whether or not the submission should be graded.
        # The grade_all function above is an example.

        self.assignment = assignment
        # TODO: Check if it exists ^

        self.submission_path = submission_path
        # TODO: Check if it exists ^

        self.filter_fn = filter_fn
        # TODO: Check if it exists ^

        self.submissions = []

        self.load_submissions()

    def has_interactive(self):
        """
        Return True if and only if grading will produce command-line
        prompts.
        """
        return self.assignment.has_interactive()

    def has_noninteractive(self):
        """
        Return True if and only if grading will produce command-line
        prompts.
        """
        return self.assignment.has_noninteractive()

    def load_submissions(self):
        self.submissions = []

        # Walk through the immediate subdirectories of the submissions path.
        subdirectories = next(os.walk(self.submission_path))[1]

        for directory in subdirectories:
            full_path = os.path.join(self.submission_path, directory)

            submission = Submission.load_from_dir(self.assignment, full_path)

            if self.filter_fn(submission):
                self.submissions.append(submission)

    def submission_count(self):
        """Return the number of submissions to grade."""
        return len(self.submissions)

    def grade(self, interactive=None):
        """
        Grade all submissions, returning an iterable of Grade instances.
        If interactive is None (the default), grade all components; if
        True, grade only interactive components, and if False, grade
        only noninteractive components.
        """

        for submission in self.submissions:
            grade = Grade(self.assignment, submission)
            grade.grade(interactive)

            yield grade

    def grades(self):
        """Return a list of all grades"""

        for submission in self.submissions:
            yield Grade(self.assignment, submission)
